import os
import math
import re
import warnings
from dataclasses import asdict, dataclass, field, replace
from enum import Enum
from typing import List, Optional, Tuple, Union

import torch
import torch.nn as nn
import torch.nn.functional as F
from tqdm import tqdm
from transformers.pytorch_utils import Conv1D

# from peft.import_utils import is_bnb_available, is_bnb_4bit_available
from peft.utils import (
    CLAMP_QUANTILE,
    COMMON_LAYERS_PATTERN,
    ModulesToSaveWrapper,
    PeftConfig,
    _freeze_adapter,
    _get_submodules,
    transpose,
)

from ..utils import DeltaTunerType, TRANSFORMERS_MODELS_TO_SSF_TARGET_MODULES_MAPPING

# if is_bnb_available():
#     import bitsandbytes as bnb

@dataclass
class SSFConfig(PeftConfig):
    """
    This is the configuration class to store the configuration of a [`SSFModel`].

    Args:
        target_modules (`Union[List[str],str]`): The names of the modules to apply ssf to.
        bias (`str`): Bias type for ssf. Can be 'none', 'all'. If 'all', the
            corresponding biases will be updated during training. Be aware that this means that, even when disabling
            the adapters, the model will not produce the same output as the base model would have without adaptation.
        layers_to_transform (`Union[List[int],int]`):
            The layer indexes to transform, if this argument is specified, it will apply the ssf transformations on
            the layer indexes that are specified in this list. If a single integer is passed, it will apply the ssf
            transformations on the layer at this index.
        layers_pattern (`str`):
            The layer pattern name, used only if `layers_to_transform` is different from `None` and if the layer
            pattern is not in the common layers pattern.
        modules_to_save (`List[str]`):List of modules apart from SSF layers to be set as trainable
            and saved in the final checkpoint.
    """

    bias: str = field(default="none", metadata={"help": "Bias type for SSF. Can be 'none', 'all'"})
    target_modules: Optional[Union[List[str], str]] = field(
        default=None,
        metadata={
            "help": "List of module names or regex expression of the module names to replace with ssf."
            "For example, ['q', 'v'] or '.*decoder.*(SelfAttention|EncDecAttention).*(q|v)$' "
        },
    )
    layers_to_transform: Optional[Union[List, int]] = field(
        default=None,
        metadata={
            "help": "The layer indexes to transform, is this argument is specified, PEFT will transform only the layers indexes that are specified inside this list. If a single integer is passed, PEFT will transform only the layer at this index."
        },
    )
    layers_pattern: Optional[str] = field(
        default=None,
        metadata={
            "help": "The layer pattern name, used only if `layers_to_transform` is different to None and if the layer pattern is not in the common layers pattern."
        },
    )
    modules_to_save: Optional[List[str]] = field(
        default=None,
        metadata={
            "help": "List of modules apart from SSF layers to be set as trainable and saved in the final checkpoint. "
            "For example, in Sequence Classification or Token Classification tasks, "
            "the final layer `classifier/score` are randomly initialized and as such need to be trainable and saved."
        },
    )

    def __post_init__(self):
        self.peft_type = DeltaTunerType.SSF

    @classmethod
    def from_pretrained(cls, pretrained_model_name_or_path: str, subfolder: Optional[str] = None, **kwargs):
        r"""
        This method loads the configuration of your adapter model from a directory.

        Args:
            pretrained_model_name_or_path (`str`):
                The directory or the Hub repository id where the configuration is saved.
            kwargs (additional keyword arguments, *optional*):
                Additional keyword arguments passed along to the child class initialization.
        """
        from peft.utils import CONFIG_NAME
        
        path = (
            os.path.join(pretrained_model_name_or_path, subfolder)
            if subfolder is not None
            else pretrained_model_name_or_path
        )

        hf_hub_download_kwargs, class_kwargs, _ = cls._split_kwargs(kwargs)

        if os.path.isfile(os.path.join(path, CONFIG_NAME)):
            config_file = os.path.join(path, CONFIG_NAME)
        else:
            try:
                config_file = hf_hub_download(
                    pretrained_model_name_or_path, CONFIG_NAME, subfolder=subfolder, **hf_hub_download_kwargs
                )
            except Exception:
                raise ValueError(f"Can't find '{CONFIG_NAME}' at '{pretrained_model_name_or_path}'")

        loaded_attributes = cls.from_json_file(config_file)

        config_cls = cls

        config = config_cls(**class_kwargs)

        for key, value in loaded_attributes.items():
            if hasattr(config, key):
                setattr(config, key, value)

        return config


class DeltaSSFSearchSpace:
    @classmethod
    def generate_search_space(cls, model, denas_config):
        search_space = {}
        search_space_name = ["num_hidden_layers"]
        for i in range(getattr(model.config, denas_config.layer_name)):
            search_space["num_hidden_layers_{}".format(i)] = [0,1]
        return search_space, search_space_name
    
    @classmethod
    def generate_supernet(cls, model, denas_config, search_space):
        supernet_config = {}
        supernet_config["num_hidden_layers"] = [search_space["num_hidden_layers_{}".format(i)][-1] for i in range(getattr(model.config, denas_config.layer_name))]
        return supernet_config

class DeltaSSFModel(torch.nn.Module):
    """
    Creates SSF model from a pretrained transformers model.

    Args:
        model ([`~transformers.PreTrainedModel`]): The model to be adapted.
        config ([`SSFConfig`]): The configuration of the SSF model.

    Returns:
        `torch.nn.Module`: The SSF model.

    **Attributes**:
        - **model** ([`~transformers.PreTrainedModel`]) -- The model to be adapted.
        - **peft_config** ([`SSFConfig`]): The configuration of the SSF model.
    """

    def __init__(self, model, config, adapter_name):
        super().__init__()
        self.model = model
        self.forward = self.model.forward
        self.peft_config = config
        self.adapter_name = adapter_name
        self.add_adapter(adapter_name, self.peft_config[adapter_name])

        # transformers models have a .config attribute, whose presence is assumed later on
        if not hasattr(self, "config"):
            self.config = {"model_type": "custom"}

    def add_adapter(self, adapter_name, config=None):
        if config is not None:
            model_config = getattr(self.model, "config", {"model_type": "custom"})
            if hasattr(model_config, "to_dict"):
                model_config = model_config.to_dict()

            config = self._prepare_ssf_config(config, model_config)
            self.peft_config[adapter_name] = config
        self._find_and_replace(adapter_name)
        mark_only_ssf_as_trainable(self.model, self.peft_config[adapter_name].bias)
        if self.peft_config[adapter_name].inference_mode:
            # _freeze_adapter(self.model, adapter_name)
            for n, p in self.model.named_parameters():
                p.requires_grad = False

    # def _check_quantization_dependency(self):
    #     loaded_in_4bit = getattr(self.model, "is_loaded_in_4bit", False)
    #     loaded_in_8bit = getattr(self.model, "is_loaded_in_8bit", False)
    #     if (loaded_in_4bit or loaded_in_8bit) and not is_bnb_available():
    #         raise ImportError(
    #             "To use SSF with 8-bit or 4-bit quantization, please install the `bitsandbytes` package. "
    #             "You can install it with `pip install bitsandbytes`."
    #         )

    def _check_target_module_exists(self, ssf_config, key):
        if isinstance(ssf_config.target_modules, str):
            target_module_found = re.fullmatch(ssf_config.target_modules, key)
        else:
            target_module_found = any(key.endswith(target_key) for target_key in ssf_config.target_modules)
            is_using_layer_indexes = getattr(ssf_config, "layers_to_transform", None) is not None
            layer_indexing_pattern = getattr(ssf_config, "layers_pattern", None)
            if is_using_layer_indexes and target_module_found:
                layers_pattern = COMMON_LAYERS_PATTERN if layer_indexing_pattern is None else layer_indexing_pattern
                layers_pattern = [layers_pattern] if isinstance(layers_pattern, str) else layers_pattern

                for pattern in layers_pattern:
                    layer_index = re.match(f".*.{pattern}\.(\d+)\.*", key)
                    if layer_index is not None:
                        layer_index = int(layer_index.group(1))
                        if isinstance(ssf_config.layers_to_transform, int):
                            target_module_found = layer_index == ssf_config.layers_to_transform
                        else:
                            target_module_found = layer_index in ssf_config.layers_to_transform

                        break
                    else:
                        target_module_found = False
        return target_module_found

    def _create_new_module(self, ssf_config, adapter_name, target):
        if isinstance(target, torch.nn.Linear):
            in_features, out_features = target.in_features, target.out_features
        else:
            raise ValueError(
                f"Target module {target} is not supported. "
                f"Currently, only `torch.nn.Linear` are supported."
            )
        new_module = Linear(target, out_features)

        return new_module

    def _find_and_replace(self, adapter_name):
        ssf_config = self.peft_config[adapter_name]
        # self._check_quantization_dependency()
        is_target_modules_in_base_model = False
        key_list = [key for key, _ in self.model.named_modules()]

        for key in key_list:
            if not self._check_target_module_exists(ssf_config, key):
                continue

            is_target_modules_in_base_model = True
            parent, target, target_name = _get_submodules(self.model, key)

            new_module = self._create_new_module(ssf_config, adapter_name, target)
            self._replace_module(parent, target_name, new_module, target)

        if not is_target_modules_in_base_model:
            raise ValueError(
                f"Target modules {ssf_config.target_modules} not found in the base model. "
                f"Please check the target modules and try again."
            )

    def _replace_module(self, parent_module, child_name, new_module, old_module):
        setattr(parent_module, child_name, new_module)
        
        # dispatch to correct device
        if hasattr(old_module, "weight"):
            for name, module in new_module.named_modules():
                module.to(old_module.weight.device)

    def _create_org_module(self, target):
        if isinstance(target, Linear):
            new_module = target.get_layer(merge=False)
        elif isinstance(target, torch.nn.Linear):
            new_module = target
        else:
            raise ValueError(
                f"Target module {target} is not supported. ")
        return new_module

    def set_sample_config(self, model_config):
        ssf_config = self.peft_config[self.adapter_name]
        # self._check_quantization_dependency()
        is_target_modules_in_base_model = False
        key_list = [key for key, _ in self.model.named_modules()]
        
        for key in key_list:
            if not self._check_target_module_exists(ssf_config, key):
                continue

            is_target_modules_in_base_model = True
            parent, target, target_name = _get_submodules(self.model, key)
            
            l_num = [int(k) for k in key.split(".") if k.isnumeric()].pop()
            if model_config["num_hidden_layers"][l_num]:
                if not isinstance(target, SSFLayer):
                    new_module = self._create_new_module(ssf_config, self.adapter_name, target)
                    self._replace_module(parent, target_name, new_module, target)
            else:
                new_module = self._create_org_module(target)
                self._replace_module(parent, target_name, new_module, target)

        if not is_target_modules_in_base_model:
            raise ValueError(
                f"Target modules {ssf_config.target_modules} not found in the base model. "
                f"Please check the target modules and try again.")
        mark_only_ssf_as_trainable(self.model, self.peft_config[self.adapter_name].bias)
        if self.peft_config[self.adapter_name].inference_mode:
            # _freeze_adapter(self.model, self.adapter_name)
             for n, p in self.model.named_parameters():
                p.requires_grad = False
        return self

    def merge_adapter(self):
        """
        This method merges the SSF layers into the base model.
        """
        for module in self.model.modules():
            if isinstance(module, SSFLayer):
                module.merge()

    def unmerge_adapter(self):
        """
        This method unmerges the SSF layers from the base model.
        """
        for module in self.model.modules():
            if isinstance(module, SSFLayer):
                module.unmerge()

    def __getattr__(self, name: str):
        """Forward missing attributes to the wrapped module."""
        try:
            return super().__getattr__(name)  # defer to nn.Module's logic
        except AttributeError:
            return getattr(self.model, name)

    def get_peft_config_as_dict(self, inference: bool = False):
        config_dict = {}
        for key, value in self.peft_config.items():
            config = {k: v.value if isinstance(v, Enum) else v for k, v in asdict(value).items()}
            if inference:
                config["inference_mode"] = True
        config_dict[key] = config
        return config

    @staticmethod
    def _prepare_ssf_config(peft_config, model_config):
        if peft_config.target_modules is None:
            if model_config["model_type"] not in TRANSFORMERS_MODELS_TO_SSF_TARGET_MODULES_MAPPING:
                raise ValueError("Please specify `target_modules` in `peft_config`")
            peft_config.target_modules = TRANSFORMERS_MODELS_TO_SSF_TARGET_MODULES_MAPPING[model_config["model_type"]]
        return peft_config

    def _unload_and_optionally_merge(self, merge=True, progressbar: bool = False):
        key_list = [key for key, _ in self.model.named_modules() if "ssf" not in key]
        desc = "Unloading " + ("and merging " if merge else "") + "model"
        for key in tqdm(key_list, disable=not progressbar, desc=desc):
            try:
                parent, target, target_name = _get_submodules(self.model, key)
            except AttributeError:
                continue
            if isinstance(target, SSFLayer):
                if isinstance(target, Linear):
                    new_module = target.get_layer(merge=merge)
                else:
                    raise NotImplementedError
                self._replace_module(parent, target_name, new_module, target)

            # save any additional trainable modules part of `modules_to_save`
            # if isinstance(target, ModulesToSaveWrapper):
            #     setattr(parent, target_name, target.modules_to_save[target.active_adapter])

        return self.model

    def merge_and_unload(self, progressbar: bool = False):
        r"""
        This method merges the SSF layers into the base model. This is needed if someone wants to use the base model
        as a standalone model.

        Args:
            progressbar (bool): whether to show a progressbar indicating the unload and merge process
        """
        return self._unload_and_optionally_merge(progressbar=progressbar)

    def unload(self):
        """
        Gets back the base model by removing all the ssf modules without merging. This gives back the original base
        model.
        """
        return self._unload_and_optionally_merge(merge=False)

# had to adapt it for `ssf_only` to work
def mark_only_ssf_as_trainable(model: nn.Module, bias: str = "none") -> None:
    for n, p in model.named_parameters():
        if "ssf_" not in n:
            p.requires_grad = False
    if bias is None or bias == "none":
        return
    elif bias == "all":
        for n, p in model.named_parameters():
            if "bias" in n:
                p.requires_grad = True
    else:
        raise NotImplementedError

class SSFLayer:
    def __init__(self, dim: int, **kwargs):
        self.dim = dim
        
        self.ssf_scale = nn.Parameter(torch.ones(dim))
        self.ssf_shift = nn.Parameter(torch.zeros(dim))
        
        nn.init.normal_(self.ssf_scale, mean=1, std=.02)
        nn.init.normal_(self.ssf_shift, std=.02)

    def ssf_ada(self, x):
        assert self.ssf_scale.shape == self.ssf_shift.shape
        if x.shape[-1] == self.ssf_scale.shape[0]:
            return x * self.ssf_scale + self.ssf_shift
        # elif x.shape[1] == self.ssf_scale.shape[0]:
        #     return x * self.ssf_scale.view(1, -1, 1, 1) + self.ssf_shift.view(1, -1, 1, 1)
        else:
            raise ValueError('the input tensor shape does not match the shape of the scale factor.')

class Linear(nn.Module, SSFLayer):
    # SSF implemented in a dense layer
    def __init__(
        self,
        target,
        out_features: int,
    ):
        super().__init__()
        SSFLayer.__init__(self, dim = out_features)
        self.linear_sub = target
        self.merged = False
        
    def merge(self):
        if self.merged:
            warnings.warn("Already merged. Nothing to do.")
            return
        
        #save unmerged data
        self.linear_sub_weight = self.linear_sub.weight.data.clone()
        self.linear_sub_bias = None if self.linear_sub.bias is None else self.linear_sub.bias.data.clone()

        # do merge operation
        self.linear_sub.weight.data = (self.linear_sub.weight.t() * self.ssf_scale).t()
        if self.linear_sub.bias is None:
            self.linear_sub.bias = self.ssf_shift
        else:
            self.linear_sub.bias.data = self.linear_sub.bias * self.ssf_scale + self.ssf_shift
        self.merged = True
        # print("Merge SSF layer into base model")
    
    def get_layer(self,merge=True):
        if merge and not self.merged:
            self.merge()
        if not merge and self.merged:
            self.unmerge()
        return self.linear_sub
    
    def unmerge(self):
        if not self.merged:
            warnings.warn("Already unmerged. Nothing to do.")
            return

        self.linear_sub.weight.data = self.linear_sub_weight
        if self.linear_sub_bias is None:
            self.linear_sub.bias = None
        else:
            self.linear_sub.bias.data = self.linear_sub_bias
        self.merged = False
        # print("Unmerge SSF layer from base model")

    def forward(self, x: torch.Tensor):
        previous_dtype = x.dtype
        if self.merged:
            result = self.linear_sub(x)
        else:
            x = self.linear_sub(x)
            result = self.ssf_ada(x)
        result = result.to(previous_dtype)
        return result



from .deltatuner_model import DeltaTunerModel
from .deltatuner import optimize
from .deltatuner_args import DeltaTunerArguments
from .mapping import DELTATUNER_TYPE_TO_CONFIG_MAPPING, get_delta_config
from .tuner import SSFConfig, DeltaSSFModel
from .utils import (
    DeltaTunerType, 
    get_deltatuner_model_state_dict, 
    set_deltatuner_model_state_dict
)

from . import scores
from . import search
from . import tuner


import torch
from transformers import pytorch_utils
from peft.tuners.lora import Linear, Conv1D, LoraLayer

LINEAR_LAYER_STRUCTURE_NAME = ["output", "intermediate", "mlp", "ffn"]

ATTN_LAYER_STRUCTURE_NAME = ["attn", "atten"]

LINEAR_MODULE_TYPE =  [torch.nn.Linear, pytorch_utils.Conv1D, Linear, Conv1D]

LINEAR_LORA_MODULE_TYPE = [LoraLayer]


import os
import torch
import json
import inspect
import random
import logging
import numpy as np
from typing import Any, Optional, Union
from peft.utils import (
    PeftType,
    get_peft_model_state_dict, 
    set_peft_model_state_dict,
    WEIGHTS_NAME,
    SAFETENSORS_WEIGHTS_NAME
)

from transformers import PreTrainedModel, AutoTokenizer
from peft import PeftModel, PeftConfig, LoraConfig, AdaLoraConfig
from peft.utils import PeftType, PromptLearningConfig, WEIGHTS_NAME, SAFETENSORS_WEIGHTS_NAME, hub_file_exists, set_peft_model_state_dict
from accelerate import dispatch_model, infer_auto_device_map
from accelerate.utils import get_balanced_memory
from accelerate.hooks import AlignDevicesHook, remove_hook_from_submodules, add_hook_to_module
from huggingface_hub import hf_hub_download
from huggingface_hub.utils import EntryNotFoundError
from safetensors.torch import load_file as safe_load_file
from safetensors.torch import save_file as safe_save_file

from .deltatuner_args import DeltaTunerArguments
from .tuner import DeltaLoraModel, DeltaLoraSearchSpace, DeltaSSFModel, DeltaSSFSearchSpace
from .search import SearchEngineFactory, Timer
from .search.utils import network_latency
from .utils import DeltaTunerType, get_deltatuner_model_state_dict, set_deltatuner_model_state_dict
from typing import Any, Dict, List, Optional, Union

DELTATUNNER_TO_MODEL_MAPPING = {
     PeftType.LORA: DeltaLoraModel,
     DeltaTunerType.SSF: DeltaSSFModel,
}

DELTATUNNER_TO_SEARCH_SPACE = {
     PeftType.LORA: DeltaLoraSearchSpace,
     DeltaTunerType.SSF: DeltaSSFSearchSpace
}

MODEL_TYPE_TO_LAYER_NAME = {
    "mpt": "n_layers"
}

def setup_seed(seed):
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)

class DeltaTunerModel(PeftModel, torch.nn.Module):
    def __init__(self, model, peft_config: PeftConfig, adapter_name: str = "default", denas_config: DeltaTunerArguments = None, tokenizer: AutoTokenizer = None):
        torch.nn.Module.__init__(self)
        self.base_model = model
        self.tokenizer = tokenizer
        self.config = getattr(self.base_model, "config", {"model_type": "custom"})
        self.modules_to_save = None
        self.best_model_param = None
        self.peft_config = {}
        self.active_adapter = adapter_name
        self.peft_type = peft_config.peft_type
        self.peft_config[adapter_name] = peft_config
        logging.basicConfig(level=logging.INFO,
                format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        self.logger = logging.getLogger('deltatuner')
        if not isinstance(model, PeftModel):
            self.base_model = DELTATUNNER_TO_MODEL_MAPPING[peft_config.peft_type](
                self.base_model, self.peft_config, adapter_name
            )
            self.set_additional_trainable_modules(peft_config, adapter_name)

            if getattr(model, "is_gradient_checkpointing", True):
                model = self._prepare_model_for_gradient_checkpointing(model)

            # the `pretraining_tp` is set for some models to simulate Tensor Parallelism during inference to avoid
            # numerical differences, https://github.com/pytorch/pytorch/issues/76232 - to avoid any unexpected
            # behavior we disable that in this line.
            if hasattr(self.base_model, "config") and hasattr(self.base_model.config, "pretraining_tp"):
                self.base_model.config.pretraining_tp = 1

        self.denas_config = denas_config
        if self.denas_config.denas:
            self._init_denas_params_()
            search_space, search_space_name = DELTATUNNER_TO_SEARCH_SPACE[peft_config.peft_type].generate_search_space(model, self.denas_config)
            supernet_config =  DELTATUNNER_TO_SEARCH_SPACE[peft_config.peft_type].generate_supernet(model, self.denas_config, search_space)
            if peft_config.peft_type in (DeltaTunerType.SSF):
                super_net = self.base_model
            else:
                super_net = DELTATUNNER_TO_MODEL_MAPPING[peft_config.peft_type](model, self.peft_config, adapter_name, supernet_config)
            if self.denas_config.best_model_structure and os.path.exists(self.denas_config.best_model_structure):
                with open(self.denas_config.best_model_structure, "r+") as best_model_param_file:
                    self.best_model_param = json.loads(best_model_param_file.readline().strip())
            else:
                self.denas_config.search_space_name = search_space_name
                self.best_model_param = json.loads(self.search(self.denas_config, super_net, search_space))
            self.base_model = super_net.set_sample_config(self.best_model_param)

    def _init_denas_params_(self):
        if self.config.model_type in MODEL_TYPE_TO_LAYER_NAME:
            self.denas_config.layer_name = MODEL_TYPE_TO_LAYER_NAME[self.config.model_type]
        self.denas_config.model_id = self.base_model.config._name_or_path
        self.denas_config.tokenizer = self.tokenizer
        self.denas_config.max_param_limits = sum(param.numel() for param in self.base_model.parameters() if param.requires_grad) / 10.**6 if self.denas_config.max_param_limits is None else self.denas_config.max_param_limits
        if self.tokenizer:
            self.denas_config.budget_latency_max = network_latency(self.base_model, self.tokenizer, batch_size=self.denas_config.batch_size) if self.denas_config.budget_latency_max is not None else self.denas_config.budget_latency_max

    def search(self, denas_config, super_net, search_space):
        setup_seed(denas_config.random_seed)
        with Timer("DE-NAS search best model"):
            searcher = SearchEngineFactory.create_search_engine(params=denas_config, super_net=super_net, search_space=search_space, peft_type=self.peft_type)
            best_structure=searcher.search()
        self.logger.info(f"DE-NAS completed, best structure is {best_structure}")
        return best_structure

    def save_pretrained(
        self,
        save_directory: str,
        safe_serialization: bool = False,
        selected_adapters: Optional[List[str]] = None,
        **kwargs: Any,
    ):
        r"""
        This function saves the adapter model and the adapter configuration files to a directory, so that it can be
        reloaded using the [`LoraModel.from_pretrained`] class method, and also used by the [`LoraModel.push_to_hub`]
        method.

        Args:
            save_directory (`str`):
                Directory where the adapter model and configuration files will be saved (will be created if it does not
                exist).
            kwargs (additional keyword arguments, *optional*):
                Additional keyword arguments passed along to the `push_to_hub` method.
        """
        if os.path.isfile(save_directory):
            raise ValueError(f"Provided path ({save_directory}) should be a directory, not a file")

        if selected_adapters is None:
            selected_adapters = list(self.peft_config.keys())
        else:
            if any(
                selected_adapter_name not in list(self.peft_config.keys())
                for selected_adapter_name in selected_adapters
            ):
                raise ValueError(
                    f"You passed an invalid `selected_adapters` arguments, current supported adapter names are"
                    f" {list(self.peft_config.keys())} - got {selected_adapters}."
                )

        os.makedirs(save_directory, exist_ok=True)
        self.create_or_update_model_card(save_directory)

        for adapter_name in selected_adapters:
            peft_config = self.peft_config[adapter_name]
            # save only the trainable weights
            if peft_config.peft_type in (DeltaTunerType.SSF):
                output_state_dict = get_deltatuner_model_state_dict(
                self, state_dict=kwargs.get("state_dict", None), adapter_name=adapter_name
            )
            else:
                output_state_dict = get_peft_model_state_dict(
                    self, state_dict=kwargs.get("state_dict", None), adapter_name=adapter_name
                )
            output_dir = os.path.join(save_directory, adapter_name) if adapter_name != "default" else save_directory
            os.makedirs(output_dir, exist_ok=True)

            if safe_serialization:
                safe_save_file(
                    output_state_dict,
                    os.path.join(output_dir, SAFETENSORS_WEIGHTS_NAME),
                    metadata={"format": "pt"},
                )
            else:
                torch.save(output_state_dict, os.path.join(output_dir, WEIGHTS_NAME))

            # save the config and change the inference mode to `True`
            if peft_config.base_model_name_or_path is None:
                peft_config.base_model_name_or_path = (
                    self.base_model.__dict__.get("name_or_path", None)
                    if isinstance(peft_config, PromptLearningConfig)
                    else self.base_model.model.__dict__.get("name_or_path", None)
                )
            inference_mode = peft_config.inference_mode
            peft_config.inference_mode = True

            if peft_config.task_type is None:
                # deal with auto mapping
                base_model_class = self._get_base_model_class(
                    is_prompt_tuning=isinstance(peft_config, PromptLearningConfig)
                )
                parent_library = base_model_class.__module__

                auto_mapping_dict = {
                    "base_model_class": base_model_class.__name__,
                    "parent_library": parent_library,
                }
            else:
                auto_mapping_dict = None

            peft_config.save_pretrained(output_dir, auto_mapping_dict=auto_mapping_dict)
            peft_config.inference_mode = inference_mode

        if self.best_model_param is not None:
            with open(os.path.join(save_directory, "best_model_structure.txt"), "w+") as fout:
                fout.write(json.dumps(self.best_model_param))

    @classmethod
    def from_pretrained(
        cls,
        model: PreTrainedModel,
        model_id: Union[str, os.PathLike],
        adapter_name: str = "default",
        is_trainable: bool = False,
        config: Optional[PeftConfig] = None,
        **kwargs: Any,
    ):
        from peft.mapping import PEFT_TYPE_TO_CONFIG_MAPPING
        from .mapping import DELTATUNER_TYPE_TO_CONFIG_MAPPING, MODEL_TYPE_TO_DELTATUNER_MODEL_MAPPING
        denas_config = kwargs.pop("denas_config", None)

        best_structure_file = os.path.join(model_id, "best_model_structure.txt")
        if os.path.isfile(best_structure_file):
            denas_config.denas = True
            denas_config.best_model_structure = best_structure_file
        else:
            denas_config.denas = False

        # load the config
        if config is None:
            peft_type_name = PeftConfig._get_peft_type(
                    model_id,
                    subfolder=kwargs.get("subfolder", None),
                    revision=kwargs.get("revision", None),
                    cache_dir=kwargs.get("cache_dir", None),
                    use_auth_token=kwargs.get("use_auth_token", None),
                )
            if peft_type_name in ['SSF']:
                config = DELTATUNER_TYPE_TO_CONFIG_MAPPING[peft_type_name].from_pretrained(model_id, **kwargs)
            else:
                config = PEFT_TYPE_TO_CONFIG_MAPPING[peft_type_name].from_pretrained(model_id, **kwargs)
        elif isinstance(config, PeftConfig):
            config.inference_mode = not is_trainable
        else:
            raise ValueError(f"The input config must be a PeftConfig, got {config.__class__}")

        if (getattr(model, "hf_device_map", None) is not None) and len(
            set(model.hf_device_map.values()).intersection({"cpu", "disk"})
        ) > 0:
            remove_hook_from_submodules(model)

        if isinstance(config, PromptLearningConfig) and is_trainable:
            raise ValueError("Cannot set a prompt learning adapter to trainable when loading pretrained adapter.")
        else:
            config.inference_mode = not is_trainable

        if config.task_type not in MODEL_TYPE_TO_DELTATUNER_MODEL_MAPPING.keys():
            model = cls(model, config, adapter_name, denas_config)
        else:
            model = MODEL_TYPE_TO_DELTATUNER_MODEL_MAPPING[config.task_type](model, config, adapter_name, denas_config)
        model.load_adapter(model_id, adapter_name, is_trainable=is_trainable, **kwargs)
        return model

    def load_adapter(self, model_id: str, adapter_name: str, is_trainable: bool = False, **kwargs: Any):
        from peft.mapping import PEFT_TYPE_TO_CONFIG_MAPPING
        from .mapping import DELTATUNER_TYPE_TO_CONFIG_MAPPING

        hf_hub_download_kwargs, kwargs = self._split_kwargs(kwargs)

        if adapter_name not in self.peft_config:
            # load the config
            peft_type_name = PeftConfig._get_peft_type(
                    model_id,
                    **hf_hub_download_kwargs,
                )
            if peft_type_name in ["SSF"]:
                peft_config = DELTATUNER_TYPE_TO_CONFIG_MAPPING[peft_type_name].from_pretrained(
                    model_id,
                    **hf_hub_download_kwargs,
                )
            else:
                peft_config = PEFT_TYPE_TO_CONFIG_MAPPING[peft_type_name].from_pretrained(
                    model_id,
                    **hf_hub_download_kwargs,
                )
            if isinstance(peft_config, PromptLearningConfig) and is_trainable:
                raise ValueError("Cannot set a prompt learning adapter to trainable when loading pretrained adapter.")
            else:
                peft_config.inference_mode = not is_trainable
            self.add_adapter(adapter_name, peft_config)

        # load weights if any
        path = (
            os.path.join(model_id, hf_hub_download_kwargs["subfolder"])
            if hf_hub_download_kwargs.get("subfolder", None) is not None
            else model_id
        )

        if os.path.exists(os.path.join(path, SAFETENSORS_WEIGHTS_NAME)):
            filename = os.path.join(path, SAFETENSORS_WEIGHTS_NAME)
            use_safetensors = True
        elif os.path.exists(os.path.join(path, WEIGHTS_NAME)):
            filename = os.path.join(path, WEIGHTS_NAME)
            use_safetensors = False
        else:
            has_remote_safetensors_file = hub_file_exists(
                model_id,
                SAFETENSORS_WEIGHTS_NAME,
                revision=hf_hub_download_kwargs.get("revision", None),
                repo_type=hf_hub_download_kwargs.get("repo_type", None),
            )
            use_safetensors = has_remote_safetensors_file

            if has_remote_safetensors_file:
                # Priority 1: load safetensors weights
                filename = hf_hub_download(
                    model_id,
                    SAFETENSORS_WEIGHTS_NAME,
                    **hf_hub_download_kwargs,
                )
            else:
                try:
                    filename = hf_hub_download(model_id, WEIGHTS_NAME, **hf_hub_download_kwargs)
                except EntryNotFoundError:
                    raise ValueError(
                        f"Can't find weights for {model_id} in {model_id} or in the Hugging Face Hub. "
                        f"Please check that the file {WEIGHTS_NAME} or {SAFETENSORS_WEIGHTS_NAME} is present at {model_id}."
                    )

        if use_safetensors:
            adapters_weights = safe_load_file(filename, device=torch.device("cuda" if torch.cuda.is_available() else "cpu"))
        else:
            adapters_weights = torch.load(filename, map_location=torch.device("cuda" if torch.cuda.is_available() else "cpu"))

        # load the weights into the model
        if self.peft_config[adapter_name].peft_type in (DeltaTunerType.SSF):
            load_result = set_deltatuner_model_state_dict(self, adapters_weights, adapter_name=adapter_name)    
        else:
            load_result = set_peft_model_state_dict(self, adapters_weights, adapter_name=adapter_name)
        if (
            (getattr(self, "hf_device_map", None) is not None)
            and (len(set(self.hf_device_map.values()).intersection({"cpu", "disk"})) > 0)
            and len(self.peft_config) == 1
        ):
            device_map = kwargs.get("device_map", "auto")
            max_memory = kwargs.get("max_memory", None)
            offload_dir = kwargs.get("offload_folder", None)
            offload_index = kwargs.get("offload_index", None)

            dispatch_model_kwargs = {}
            # Safety checker for previous `accelerate` versions
            # `offload_index` was introduced in https://github.com/huggingface/accelerate/pull/873/
            if "offload_index" in inspect.signature(dispatch_model).parameters:
                dispatch_model_kwargs["offload_index"] = offload_index

            no_split_module_classes = self._no_split_modules

            if device_map != "sequential":
                max_memory = get_balanced_memory(
                    self,
                    max_memory=max_memory,
                    no_split_module_classes=no_split_module_classes,
                    low_zero=(device_map == "balanced_low_0"),
                )
            if isinstance(device_map, str):
                device_map = infer_auto_device_map(
                    self, max_memory=max_memory, no_split_module_classes=no_split_module_classes
                )
            dispatch_model(
                self,
                device_map=device_map,
                offload_dir=offload_dir,
                **dispatch_model_kwargs,
            )
            hook = AlignDevicesHook(io_same_device=True)
            if isinstance(self.peft_config[adapter_name], PromptLearningConfig):
                remove_hook_from_submodules(self.prompt_encoder)
            add_hook_to_module(self.get_base_model(), hook)

        # Set model in evaluation mode to deactivate Dropout modules by default
        if not is_trainable:
            self.eval()
        return load_result

class DelatunerModelForCausalLM(DeltaTunerModel):
    def __init__(self, model: PeftModel, peft_config: PeftConfig, adapter_name: str = "default", denas_config: DeltaTunerArguments = None, tokenizer: AutoTokenizer = None):
        super().__init__(model, peft_config, adapter_name, denas_config, tokenizer)
        self.base_model_prepare_inputs_for_generation = self.base_model.prepare_inputs_for_generation
    
    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        inputs_embeds=None,
        labels=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
        **kwargs,
    ):
        peft_config = self.active_peft_config
        if self.base_model.config.model_type == "mpt":
            if inputs_embeds is not None:
                raise AssertionError("forward in MPTForCausalLM does not support inputs_embeds")
            return self.base_model(
                input_ids=input_ids,
                attention_mask=attention_mask,
                labels=labels,
                output_attentions=output_attentions,
                output_hidden_states=output_hidden_states,
                return_dict=return_dict,
                **kwargs,
            )

        return self.base_model(
            input_ids=input_ids,
            attention_mask=attention_mask,
            inputs_embeds=inputs_embeds,
            labels=labels,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
            **kwargs,
        )

    def generate(self, **kwargs):
        self.base_model.prepare_inputs_for_generation = self.prepare_inputs_for_generation
        if hasattr(self.base_model, "model"):
            self.base_model.model.generation_config = self.generation_config
        else:
            self.base_model.generation_config = self.generation_config
        try:
            outputs = self.base_model.generate(**kwargs)
        except:
            self.base_model.prepare_inputs_for_generation = self.base_model_prepare_inputs_for_generation
            raise
        else:
            self.base_model.prepare_inputs_for_generation = self.base_model_prepare_inputs_for_generation
            return outputs

    def prepare_inputs_for_generation(self, *args, **kwargs):
        peft_config = self.active_peft_config
        model_kwargs = self.base_model_prepare_inputs_for_generation(*args, **kwargs)
        
        return model_kwargs
        


# coding=utf-8
# Copyright 2023-present the HuggingFace Inc. team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .config import DeltaTunerType

def get_deltatuner_model_state_dict(model, state_dict=None, adapter_name="default"):
    """
    Get the state dict of the DeltaTuner model.

    Args:
        model ([`DeltaTunerModelModel`]): The DeltaTunerModel model. When using torch.nn.DistributedDataParallel, DeepSpeed or FSDP,
        the model should be the underlying model/unwrapped model (i.e. model.module).
        state_dict (`dict`, *optional*, defaults to `None`):
            The state dict of the model. If not provided, the state dict of the model
        will be used.
    """
    config = model.peft_config[adapter_name]
    if state_dict is None:
        state_dict = model.state_dict()
    
    if config.peft_type == DeltaTunerType.SSF:
        bias = config.bias
        if bias is None or bias == "none":
            to_return = {k: state_dict[k] for k in state_dict if "ssf_" in k}
        elif bias == "all":
            to_return = {k: state_dict[k] for k in state_dict if "ssf_" in k or "bias" in k}
        else:
            raise NotImplementedError
    else:
        raise NotImplementedError
    if model.modules_to_save is not None:
        for key, value in state_dict.items():
            if any(f"{module_name}.modules_to_save.{adapter_name}" in key for module_name in model.modules_to_save):
                to_return[key.replace("modules_to_save.", "")] = value

    to_return = {k.replace(f".{adapter_name}", ""): v for k, v in to_return.items()}
    return to_return


def set_deltatuner_model_state_dict(model, peft_model_state_dict, adapter_name="default"):
    """
    Set the state dict of the DeltaTuner model.

    Args:
        model ([`DeltaTunerModel`]): The DeltaTuner model.
        peft_model_state_dict (`dict`): The state dict of the DeltaTuner model.
    """
    config = model.peft_config[adapter_name]
    state_dict = {}
    if model.modules_to_save is not None:
        for key, value in peft_model_state_dict.items():
            if any(module_name in key for module_name in model.modules_to_save):
                for module_name in model.modules_to_save:
                    if module_name in key:
                        key = key.replace(module_name, f"{module_name}.modules_to_save.{adapter_name}")
                        break
            state_dict[key] = value
    else:
        state_dict = peft_model_state_dict

    if config.peft_type in (DeltaTunerType.SSF):
        peft_model_state_dict = {}
        for k, v in state_dict.items():
            peft_model_state_dict[k] = v
    else:
        raise NotImplementedError

    load_result = model.load_state_dict(peft_model_state_dict, strict=False)
    return load_result



from .EvolutionarySearchEngine import EvolutionarySearchEngine

SEARCHER_TYPES = {
        "EvolutionarySearchEngine": EvolutionarySearchEngine
}

class SearchEngineFactory(object):

    @staticmethod
    def create_search_engine(params, super_net, search_space,peft_type):
        if params.search_engine in SEARCHER_TYPES:
            return SEARCHER_TYPES[params.search_engine](params, super_net, search_space,peft_type)
        else:
            raise RuntimeError(f"Search Engine {params.search_engine} is not supported")


import logging
from peft import LoraConfig, PeftModel, LoraModel
from .tuner import SSFConfig

from .deltatuner_model import DeltaTunerModel
from .deltatuner_args import DeltaTunerArguments
from .mapping import MODEL_TYPE_TO_DELTATUNER_MODEL_MAPPING

logging.basicConfig(level=logging.INFO,
                format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
logger = logging.getLogger('deltatuner')

SUPPORTED_ALGO = ["auto", "lora", "ssf"]

def optimize(model, tokenizer, adapter_name: str="default", deltatuning_args: DeltaTunerArguments=None, **kwargs) -> DeltaTunerModel:
    if deltatuning_args is None:
        deltatuning_args = DeltaTunerArguments(**kwargs)

    algo = deltatuning_args.algo
    if algo not in SUPPORTED_ALGO:
        raise NotImplementedError("Current algorithm {} is not supported in deltatuner. ".format(algo))
    
    if algo == "auto":
        if "mpt" in model.config.model_type.lower() or not isinstance(model, PeftModel):
            algo = "ssf"
        else:
            algo = "lora"

    if algo == "ssf" and not isinstance(model, PeftModel):
        peft_config = SSFConfig(target_modules=deltatuning_args.ssf_target_module, bias=None, task_type=deltatuning_args.task_type)
        peft_config.base_model_name_or_path = model.__dict__.get("name_or_path", None)
    elif algo == "lora" and isinstance(model, PeftModel) and isinstance(model.peft_config[adapter_name], LoraConfig):
        peft_config = model.peft_config[adapter_name]
        if not deltatuning_args.denas:
            return model
    else:
        raise NotImplementedError("Current algorithm {} does not support {} type of model in deltatuner. Please specify the right algo with the type of model.".format(algo, type(model)))

    whether_denas = "enable" if deltatuning_args.denas else "disable"
    logging.info("The user specifies or automatically adjasts {} algo on model {} with {} denas".format(algo, model.config.model_type, whether_denas))

    if isinstance(peft_config, LoraConfig) or isinstance(peft_config, SSFConfig):
        if peft_config.task_type not in MODEL_TYPE_TO_DELTATUNER_MODEL_MAPPING:
            model = DeltaTunerModel(model, peft_config, adapter_name, deltatuning_args, tokenizer)
        else:
            model = MODEL_TYPE_TO_DELTATUNER_MODEL_MAPPING[peft_config.task_type](model, peft_config, adapter_name, deltatuning_args, tokenizer)
        return model
    else:
        raise NotImplementedError("Current algorithm {} is not supported in deltatuner. ".format(algo))




import os, sys, time, logging
import torch
from torch import nn
import numpy as np
import gc

import torch

from ..utils import DeltaTunerType
from ..search.utils import network_latency, input_constructor

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
logger = logging.getLogger('DENAS')

def do_compute_nas_score_transformer(model, tokenizer, resolution, batch_size, expressivity_weight=0, complexity_weight=0, diversity_weight=0, saliency_weight=0, latency_weight=0,peft_type=None):
    if peft_type in (DeltaTunerType.SSF):
        from .transformer_proxy_ssf import compute_diversity_score, compute_saliency_score
    else:
        from .transformer_proxy import network_weight_gaussian_init, compute_diversity_score, compute_saliency_score
        network_weight_gaussian_init(model)
    expressivity_score = 0
    complexity_score = 0
    model.train()
    #model.requires_grad_(True)
    model.zero_grad() 
    
    input, _ = input_constructor(batch_size=batch_size, resolution=resolution, tokenizer=tokenizer)
    output = model(**input)
    output = output.logits
    torch.sum(output).backward()

    diversity_score_list = compute_diversity_score(model)
    diversity_score = 0
    for grad_abs in diversity_score_list:
        if len(grad_abs.shape) == 0:
            diversity_score += float(grad_abs)
        elif len(grad_abs.shape) == 2:
            diversity_score += float(torch.mean(torch.sum(grad_abs, dim=[1])))
        else:
            raise RuntimeError("diversity_score only support grad shapre of 0 or 2")

    grads_abs_list = compute_saliency_score(model)
    saliency_score = 0
    for grad_abs in grads_abs_list:
        if len(grad_abs.shape) == 4:
            saliency_score += float(torch.mean(torch.sum(grad_abs, dim=[1,2,3])))
        elif len(grad_abs.shape) == 2:
            saliency_score += float(torch.mean(torch.sum(grad_abs, dim=[1])))
        elif len(grad_abs.shape) == 1:
            saliency_score += float(torch.mean(grad_abs))
        else:
            raise RuntimeError('saliency_score only support grad shape of 1, 2 or 4')
    
    latency = 0
    if latency_weight != 0:
        latency = network_latency(model=model, tokenizer=tokenizer, batch_size=batch_size)
    saliency_score_scale = 10.**5 if peft_type in (DeltaTunerType.SSF) else 10.**9
    score = (expressivity_score*expressivity_weight / 10.**9
                    + complexity_score*complexity_weight / 10.**9
                    + diversity_score*diversity_weight / 10.**9
                    + saliency_score*saliency_weight / saliency_score_scale)
    logger.info("saliency score:{}, diversity score:{}, latency_score:{}".format(saliency_score/saliency_score_scale, diversity_score/10.**9, latency))
    nas_score = score/(1 + latency*latency_weight)
    return nas_score, score, latency

def do_compute_nas_score(model, tokenizer, resolution, batch_size, mixup_gamma, expressivity_weight=0, complexity_weight=0, diversity_weight=0, saliency_weight=0, latency_weight=0,peft_type=None):
    nas_score, score, latency = do_compute_nas_score_transformer(model, tokenizer, resolution, batch_size, expressivity_weight, complexity_weight, diversity_weight, saliency_weight, latency_weight,peft_type)
    return nas_score, score, latency


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("deltatuner/version", "r") as fh:
    VERSION = fh.read().strip()

REQUIRED_PACKAGES = [
    'torch>=1.13.0', 'transformers', 'datasets', 'sentencepiece', 
    'peft==0.4.0', 'evaluate', 'nltk', 'rouge_score', 'einops', 
    'sigopt', 'torchsummary'
]

setuptools.setup(
    name="deltatuner",
    version=VERSION,
    author="Intel AIA",
    author_email="bdf.aiok@intel.com",
    description="Intel extension for peft with PyTorch and DENAS",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/intel/e2eAIOK/",
    download_url='https://github.com/intel/e2eAIOK/',
    packages=setuptools.find_packages(
        exclude=["example", "docker", ]),
    package_data={'deltatuner': ['version']}, 
    python_requires=">=3.7",  # '>=3.4',  # !=3.4.*
    install_requires=REQUIRED_PACKAGES,
    extras_require={

    },
    entry_points={
    },
    classifiers=(
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ),
    license="Apache-2.0",
    keywords=[
        'deep learning', 'LLM', 'fine-tuning', 'pytorch', 'peft',
        'lora', 'NAS'
    ],
)


import enum
from dataclasses import  field
from typing import List, Optional, Tuple, Union
from peft.utils import TRANSFORMERS_MODELS_TO_LORA_TARGET_MODULES_MAPPING

class DeltaTunerType(str, enum.Enum):
    SSF = "SSF"

TRANSFORMERS_MODELS_TO_SSF_TARGET_MODULES_MAPPING = TRANSFORMERS_MODELS_TO_LORA_TARGET_MODULES_MAPPING
TRANSFORMERS_MODELS_TO_SSF_TARGET_MODULES_MAPPING.update({
    "llama": ["q_proj", "v_proj"],
    "mpt": ["Wqkv","out_proj","up_proj","down_proj"]
})